# -*- coding: utf-8 -*-
#
# Clone Scanner, version 1.3 for WeeChat version 0.3
# Latest development version: https://github.com/FiXato/weechat_scripts
#
#   A Clone Scanner that can manually scan channels and
#   automatically scans joins for users on the channel
#   with multiple nicknames from the same host.
#
#   Upon join by a user, the user's host is compared to the infolist of
#   already connected users to see if they are already online from
#   another nickname. If the user is a clone, it will report it.
#   With the '/clone_scanner scan' command you can manually scan a chan.
#
#   See /set plugins.var.python.clone_scanner.* for all possible options
#   Use the brilliant iset.pl plugin (/weeget install iset) to see what they do
#   Or check the sourcecode below.
#
# Example output for an on-join scan result:
#   21:32:46  ▬▬▶ FiXato_Odie (FiXato@FiXato.net) has joined #lounge
#   21:32:46      FiXato_Odie is already on the channel as FiXato!FiXato@FiXato.Net and FiX!FiXaphone@FiXato.net
#
# Example output for a manual scan:
#   21:34:44 fixato.net is online from 3 nicks:
#   21:34:44  - FiXato!FiXato@FiXato.Net
#   21:34:44  - FiX!FiXaphone@FiXato.net
#   21:34:44  - FiXato_Odie!FiXato@FiXato.net
#
## History:
### 2011-09-11: FiXato:
#
# * version 0.1: initial release.
#     * Added an on-join clone scan. Any user that joins a channel will be
#       matched against users already on the channel.
#
# * version 0.2: manual clone scan
#     * Added a manual clone scan via /clone_scanner scan
#        you can specify a target channel with:
#         /clone_scanner scan #myChannelOnCurrentServer
#        or:
#         /clone_scanner scan Freenode.#myChanOnSpecifiedNetwork
#     * Added completion
#
### 2011-09-12: FiXato:
#
# * version 0.3: Refactor galore
#     * Refactored some code. Codebase should be DRYer and clearer now.
#     * Manual scan report lists by host instead of nick now.
#     * Case-insensitive host-matching
#     * Bugfixed the infolist memleak.
#     * on-join scanner works again
#     * Output examples added to the comments
#
### 2011-09-19
# * version 0.4: Option galore
#     * Case-insensitive buffer lookup fix.
#     * Made most messages optional through settings.
#     * Made on-join alert and clone report key a bit more configurable.
#     * Added formatting options for on-join alerts.
#     * Added format_message helper method that accepts multiple whitespace-separated weechat.color() options.
#     * Added formatting options for join messages
#     * Added formatting options for clone reports
#     * Added format_from_config helper method that reads the given formatting key from the config
#
# * version 0.5: cs_buffer refactoring
#     * dropping the manual cs_create_buffer call in favor for a cs_get_buffer() method
#
### 2012-02-10: FiXato:
#
# * version 0.6: Stop shoving that buffer in my face!
#     * The clone_scanner buffer should no longer pop up by itself when you load the script.
#       It should only pop up now when you actually a line needs to show up in the buffer.
#
# * version 0.7: .. but please pop it up in my current window when I ask for it
#     * Added setting plugins.var.python.clone_scanner.autofocus
#       This will autofocus the clone_scanner buffer in the current window if another window isn't
#       already showing it, and of course only when the clone_scanner buffer is triggered
#
### 2012-02-10: FiXato:
#
# * version 0.8: .. and only when it is first created..
#     * Prevents the buffer from being focused every time there is activity in it and not being shown in a window.
#
### 2012-04-01: FiXato:
#
# * version 0.9: Hurrah for bouncers...
#     * Added the option plugins.var.python.clone_scanner.compare_idents
#       Set it to 'on' if you don't want people with different idents to be marked as clones.
#       Useful on channels with bouncers.
#
### 2012-04-02: FiXato:
#
# * version 1.0: Bugfix
#     * Fixed the on-join scanner bug introduced by the 0.9 release.
#       I was not properly comparing the new ident@host.name key in all places yet.
#       Should really have tested this better ><
#
### 2012-04-03: FiXato:
#
# * version 1.1: Stop being so sensitive!
#     * Continuing to fix the on-join scanner bugs introduced by the 0.9 release.
#       The ident@host.name dict key wasn't being lowercased for comparison in the on-join scan.
#
# * version 1.2: So shameless!
#     * Added shameless advertising for my script through /clone_scanner advertise
#
### 2013-04-09: FiXato:
# * version 1.3: Such a killer rabbit
#     * Thanks to Curtis Sorensen aka killerrabbit clone_scanner.py now supports:
#       * local channels (&-prefixed)
#       * nameless channels (just # or &)
#
### 2014-12-07: FiXato:
# * version 1.4: Inefficiency Warning Patch
# WARNING! I recently noticed the  clone_scanner script is currently rather inefficient, and requires a rewrite.
#   It may cause nasty lag when there are a lot of concurrent joins on the channel, as it evaluates the nicklist on every join.
#   For servers like twitch.tv and bitlbee, you might want to exclude the server with the new setting:
#    /set plugins.var.python.clone_scanner.hooks.excluded_servers twitchtv,bitlbee
#   This script update is an emergency update to add the above option, and warn the users of this script.
#   You can disable this warning with: /set plugins.var.python.clone_scanner.lag_warning off
#
# * Other patches in this version include:
#     * Re-did how settings are handled, using nils_2's skeleton code
#     * Settings are now automatically memoized when changed
#     * Added options hooks.excluded_servers, hooks.explicit_servers and lag_warning
#     * Updated advertise option to include /script install clone_scanner.py
#     * Updated min version to 0.3.6, though this will prob change in the next version
#
### 2024-08-18: Sébastien Helleu:
# * version 1.5: make script compatible with Python 3
#
## Acknowledgements:
# * Sebastien "Flashcode" Helleu, for developing the kick-ass chat/IRC
#    client WeeChat
# * ArZa, whose kickban.pl script helped me get started with using the
#   infolist results.
# * LayBot, for requesting the ident comparison
# * Curtis "killerrabbit" Sorensen, for sending in two pull-requests,
#   adding support for local and nameless channels.
# * nils_2 aka weechatter, for providing the excellent skeleton.py script
#
## TODO:
#   - REWRITE TO IMPROVE EFFICIENCY:
#     - Probably only do the infolist loop on self-join,
#       and from then on manually keep track of join/parts/quit/nick-changes
#   - Add option to enable/disable public clone reporting aka msg channels
#   - Add option to enable/disable scanning on certain channels
#   - Add cross-channel clone scan
#   - Add cross-server clone scan
#
## Copyright (c) 2011-2014 Filip H.F. "FiXato" Slagter,
#   <FiXato [at] Gmail [dot] com>
#   http://profile.fixato.org
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
SCRIPT_NAME     = "clone_scanner"
SCRIPT_AUTHOR   = "Filip H.F. 'FiXato' Slagter <fixato [at] gmail [dot] com>"
SCRIPT_VERSION  = "1.5"
SCRIPT_LICENSE  = "MIT"
SCRIPT_DESC     = "A Clone Scanner that can manually scan channels and automatically scans joins for users on the channel with multiple nicknames from the same host."
SCRIPT_COMMAND  = "clone_scanner"
SCRIPT_CLOSE_CB = "cs_close_cb"

try:
  import weechat
  IMPORT_OK = True
except ImportError:
  print("This script must be run under WeeChat.")
  IMPORT_OK = False

import re
cs_buffer = None
OPTIONS = {
  "autofocus":                           ("on", "Focus the clone_scanner buffer in the current window if it isn't already displayed by a window."),
  "compare_idents":                      ("off", "Match against ident@host.name instead of just the hostname. Useful if you don't want different people from bouncers marked as clones"),
  "display_join_messages":               ("off", "Display all joins in the clone_scanner buffer"),
  "display_onjoin_alert_clone_buffer":   ("on", "Display an on-join clone alert in the clone_scanner buffer"),
  "display_onjoin_alert_target_buffer":  ("on", "Display an on-join clone alert in the buffer where the clone was detected"),
  "display_onjoin_alert_current_buffer": ("off", "Display an on-join clone alert in the current buffer"),
  "display_scan_report_clone_buffer":    ("on", "Display manual scan reports in the clone buffer"),
  "display_scan_report_target_buffer":   ("off", "Display manual scan reports in the buffer of the scanned channel"),
  "display_scan_report_current_buffer":  ("on", "Display manual scan reports in the current buffer"),

  "clone_report_key":                    ("mask", "Which 'key' to display in the clone report: 'mask' for full hostmasks, or 'nick' for nicks"),
  "clone_onjoin_alert_key":              ("mask", "Which 'key' to display in the on-join alerts: 'mask' for full hostmasks, or 'nick' for nicks"),

  "colors.onjoin_alert.message":   ("red", "The on-join clone alert's message colour. Formats are space separated."),
  "colors.onjoin_alert.nick":      ("bold red", "The on-join clone alert's nick colour. Formats are space separated. Note: if you have colorize_nicks, this option might not work as expected."),
  "colors.onjoin_alert.channel":   ("red", "The on-join clone alert's channel colour. Formats are space separated."),
  "colors.onjoin_alert.matches":   ("bold red", "The on-join clone alert's matches (masks or nicks) colour. Formats are space separated. Note: if you have colorize_nicks, this option might not work as expected."),

  "colors.join_messages.message":    ("chat", "The base colour for the join messages."),
  "colors.join_messages.nick":       ("bold", "The colour for the 'nick'-part of the join messages. Note: if you have colorize_nicks, this option might not always work as expected."),
  "colors.join_messages.identhost":  ("chat", "The colour for the 'ident@host'-part of the join messages."),
  "colors.join_messages.channel":    ("bold", "The colour for the 'channel'-part of the join messages."),

  "colors.clone_report.header.message":          ("chat", "The colour of the clone report header."),
  "colors.clone_report.header.number_of_hosts":  ("bold", "The colour of the number of hosts in the clone report header."),
  "colors.clone_report.header.channel":          ("bold", "The colour of the channel name in the clone report header."),

  "colors.clone_report.subheader.message":        ("chat", "The colour of the clone report subheader."),
  "colors.clone_report.subheader.host":              ("bold", "The colour of the host in the clone report subheader."),
  "colors.clone_report.subheader.number_of_clones":  ("bold", "The colour of the number of clones in the clone report subheader."),

  "colors.clone_report.clone.message": ("chat", "The colour of the clone hit in the clone report message."),
  "colors.clone_report.clone.match":   ("chat", "The colour of the match details (masks or nicks) in the clone report."),

  "colors.mask.nick":      ("bold", "The formatting of the nick in the match mask."),
  "colors.mask.identhost": ("", "The formatting of the identhost in the match mask."),
  "hooks.explicit_servers": ("*", "Comma-separated, wildcard-supporting list of servers for which we should add hook to for monitoring clones. E.g. 'freenode,chat4all,esper*' or '*' (default)"),
  "hooks.excluded_servers": ("bitlbee,twitchtv", "Which servers should the hook *not* be valid for? There's no support for wildcards unfortunately. E.g.: 'bitlbee,twitchtv' to exclude servers named bitlbee and twitchtv (default)."),
  "lag_warning": ('on', 'Show temporary warning upon script load regarding the inefficiency of the script. Set to "off" to disable.')
}
hooks = set([])

def get_validated_key_from_config(setting):
  key = OPTIONS[setting]
  if key != 'mask' and key != 'nick':
    weechat.prnt("", "Key %s not found. Valid settings are 'nick' and 'mask'. Reverted the setting to 'mask'" % key)
    weechat.config_set_plugin("clone_report_key", "mask")
    key = "mask"
  return key

def format_message(msg, formats, reset_color='chat'):
  if type(formats) == str:
    formats = formats.split()
  formatted_message = msg
  needs_color_reset = False
  for format in formats:
    if format in ['bold', 'reverse', 'italic', 'underline']:
      end_format = '-%s' % format
    else:
      needs_color_reset = True
      end_format = ""
    formatted_message = "%s%s%s" % (weechat.color(format), formatted_message, weechat.color(end_format))
  if needs_color_reset:
    formatted_message += weechat.color(reset_color)
  return formatted_message

def format_from_config(msg, config_option):
  return format_message(msg, OPTIONS[config_option])

def on_join_scan_cb(data, signal, signal_data):
  network = signal.split(',')[0]
  if network in OPTIONS['hooks.excluded_servers'].split(','):
    return weechat.WEECHAT_RC_OK

  joined_nick = weechat.info_get("irc_nick_from_host", signal_data)
  join_match_data = re.match(r':[^!]+!([^@]+@(\S+)) JOIN :?([#&]\S*)', signal_data)
  parsed_ident_host = join_match_data.group(1).lower()
  parsed_host = join_match_data.group(2).lower()
  if OPTIONS["compare_idents"] == "on":
    hostkey = parsed_ident_host
  else:
    hostkey = parsed_host

  chan_name = join_match_data.group(3)
  network_chan_name = "%s.%s" % (network, chan_name)
  chan_buffer = weechat.info_get("irc_buffer", "%s,%s" % (network, chan_name))
  if not chan_buffer:
    weechat.prnt("", "No IRC channel buffer found for %s" % network_chan_name)
    return weechat.WEECHAT_RC_OK

  if OPTIONS["display_join_messages"] == "on":
    message = "%s%s%s%s%s" % (
      format_from_config(joined_nick, "colors.join_messages.nick"),
      format_from_config("!", "colors.join_messages.message"),
      format_from_config(parsed_ident_host, "colors.join_messages.identhost"),
      format_from_config(" JOINed ", "colors.join_messages.message"),
      format_from_config(network_chan_name, "colors.join_messages.channel"),
    )
    #Make sure message format is also applied if no formatting is given for nick
    message = format_from_config(message, "colors.join_messages.message")
    weechat.prnt(cs_get_buffer(), message)

  clones = get_clones_for_buffer("%s,%s" % (network, chan_name), hostkey)
  if clones:
    key = get_validated_key_from_config("clone_onjoin_alert_key")

    filtered_clones = [clone for clone in clones[hostkey] if clone['nick'] != joined_nick]
    match_strings = [format_from_config(m[key], "colors.onjoin_alert.matches") for m in filtered_clones]

    join_string = format_from_config(' and ',"colors.onjoin_alert.message")
    masks = join_string.join(match_strings)
    message = "%s %s %s %s %s" % (
      format_from_config(joined_nick, "colors.onjoin_alert.nick"),
      format_from_config("is already on", "colors.onjoin_alert.message"),
      format_from_config(network_chan_name, "colors.onjoin_alert.channel"),
      format_from_config("as", "colors.onjoin_alert.message"),
      masks
    )
    message = format_from_config(message, 'colors.onjoin_alert.message')

    if OPTIONS["display_onjoin_alert_clone_buffer"] == "on":
      weechat.prnt(cs_get_buffer(),message)
    if OPTIONS["display_onjoin_alert_target_buffer"] == "on":
      weechat.prnt(chan_buffer, message)
    if OPTIONS["display_onjoin_alert_current_buffer"] == "on":
      weechat.prnt(weechat.current_buffer(),message)
  return weechat.WEECHAT_RC_OK

def cs_get_buffer():
  global cs_buffer

  if not cs_buffer:
    # Sets notify to 0 as this buffer does not need to be in hotlist.
    cs_buffer = weechat.buffer_new("clone_scanner", "", \
                "", SCRIPT_CLOSE_CB, "")
    weechat.buffer_set(cs_buffer, "title", "Clone Scanner")
    weechat.buffer_set(cs_buffer, "notify", "0")
    weechat.buffer_set(cs_buffer, "nicklist", "0")
    if OPTIONS["autofocus"] == "on":
      if not weechat.window_search_with_buffer(cs_buffer):
        weechat.command("", "/buffer " + weechat.buffer_get_string(cs_buffer,"name"))

  return cs_buffer

def cs_close_cb(*kwargs):
  """ A callback for buffer closing. """
  global cs_buffer

  #TODO: Ensure the clone_scanner buffer gets closed if its option is set and the script unloads

  cs_buffer = None
  return weechat.WEECHAT_RC_OK


def get_channel_from_buffer_args(buffer, args):
  server_name = weechat.buffer_get_string(buffer, "localvar_server")
  channel_name = args
  if not channel_name:
    channel_name = weechat.buffer_get_string(buffer, "localvar_channel")

  match_data = re.match(r'\A(irc.)?([^.]+)\.([#&]\S*)\Z', channel_name)
  if match_data:
    channel_name = match_data.group(3)
    server_name = match_data.group(2)

  return server_name, channel_name

#TODO: track the hosts + nicks ourselves instead of looking up the entire list every join...
def get_clones_for_buffer(infolist_buffer_name, hostname_to_match=None):
  matches = {}
  infolist = weechat.infolist_get("irc_nick", "", infolist_buffer_name)
  while(weechat.infolist_next(infolist)):
    ident_hostname = weechat.infolist_string(infolist, "host")
    host_matchdata = re.match(r'([^@]+)@(\S+)', ident_hostname)
    if not host_matchdata:
      continue

    hostname = host_matchdata.group(2).lower()
    ident = host_matchdata.group(1).lower()
    if OPTIONS["compare_idents"] == "on":
      hostkey = ident_hostname.lower()
    else:
      hostkey = hostname

    if hostname_to_match and hostname_to_match.lower() != hostkey:
      continue

    nick = weechat.infolist_string(infolist, "name")

    matches.setdefault(hostkey,[]).append({
      'nick': nick,
      'mask': "%s!%s" % (
        format_from_config(nick, "colors.mask.nick"),
        format_from_config(ident_hostname, "colors.mask.identhost")),
      'ident': ident,
      'ident_hostname': ident_hostname,
      'hostname': hostname,
    })
  weechat.infolist_free(infolist)

  #Select only the results that have more than 1 match for a host
  return dict((k, v) for (k, v) in matches.items() if len(v) > 1)

def report_clones(clones, scanned_buffer_name, target_buffer=None):
  # Default to clone_scanner buffer
  if not target_buffer:
    target_buffer = cs_get_buffer()

  if clones:
    clone_report_header = "%s %s %s%s" % (
      format_from_config(len(clones), "colors.clone_report.header.number_of_hosts"),
      format_from_config("hosts with clones were found on", "colors.clone_report.header.message"),
      format_from_config(scanned_buffer_name, "colors.clone_report.header.channel"),
      format_from_config(":", "colors.clone_report.header.message"),
    )
    clone_report_header = format_from_config(clone_report_header, "colors.clone_report.header.message")
    weechat.prnt(target_buffer, clone_report_header)

    for (host, clones) in clones.items():
      host_message = "%s %s %s %s" % (
        format_from_config(host, "colors.clone_report.subheader.host"),
        format_from_config("is online from", "colors.clone_report.subheader.message"),
        format_from_config(len(clones), "colors.clone_report.subheader.number_of_clones"),
        format_from_config("nicks:", "colors.clone_report.subheader.message"),
      )
      host_message = format_from_config(host_message, "colors.clone_report.subheader.message")
      weechat.prnt(target_buffer, host_message)

      for user in clones:
        key = get_validated_key_from_config("clone_report_key")
        clone_message = "%s%s" % (" - ", format_from_config(user[key], "colors.clone_report.clone.match"))
        clone_message = format_from_config(clone_message,"colors.clone_report.clone.message")
        weechat.prnt(target_buffer, clone_message)
  else:
    weechat.prnt(target_buffer, "No clones found on %s" % scanned_buffer_name)

def cs_command_main(data, buffer, args):
  if args[0:4] == 'scan':
    server_name, channel_name = get_channel_from_buffer_args(buffer, args[5:])
    clones = get_clones_for_buffer('%s,%s' % (server_name, channel_name))
    if OPTIONS["display_scan_report_target_buffer"] == "on":
      target_buffer = weechat.info_get("irc_buffer", "%s,%s" % (server_name, channel_name))
      report_clones(clones, '%s.%s' % (server_name, channel_name), target_buffer)
    if OPTIONS["display_scan_report_clone_buffer"] == "on":
      report_clones(clones, '%s.%s' % (server_name, channel_name))
    if OPTIONS["display_scan_report_current_buffer"] == "on":
      report_clones(clones, '%s.%s' % (server_name, channel_name), weechat.current_buffer())
  elif args[0:9] == 'advertise':
    weechat.command("", "/input insert /me is using FiXato's CloneScanner v%s for WeeChat. Get the latest version from: https://github.com/FiXato/weechat_scripts/blob/master/clone_scanner.py or /script install clone_scanner.py" % SCRIPT_VERSION)
  return weechat.WEECHAT_RC_OK

def cs_set_default_settings():
  global OPTIONS

  # Set default settings
  for option,value in OPTIONS.items():
    if not weechat.config_is_set_plugin(option):
        weechat.config_set_plugin(option, value[0])
        OPTIONS[option] = value[0]
    else:
        OPTIONS[option] = weechat.config_get_plugin(option)
    weechat.config_set_desc_plugin(option, '%s (default: "%s")' % (value[1], value[0]))

def toggle_refresh(pointer, name, value):
  global OPTIONS

  config_option          = name[len('plugins.var.python.' + SCRIPT_NAME + '.'):]  # get optionname
  OPTIONS[config_option] = value                                                  # save new value
  if config_option in ('hooks.excluded_servers', 'hooks.explicit_servers'):
    remove_hooks()
    add_hooks()
  weechat.config_set_plugin(config_option, value)
  return weechat.WEECHAT_RC_OK

def add_hooks():
  global hooks
  hooked_servers = OPTIONS['hooks.explicit_servers'].split(',')
  for server_name in hooked_servers:
    signal = "%s,irc_in2_join" % server_name
    # weechat.prnt('', "Adding hook on %s" % signal)
    hook = weechat.hook_signal(signal, "on_join_scan_cb", "")
    hooks.add(hook)

def remove_hooks():
  global hooks
  for hook in hooks:
    weechat.unhook(hook)
  hooks = set([])

if __name__ == "__main__" and IMPORT_OK:
  if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, SCRIPT_CLOSE_CB, ""):
    version = weechat.info_get("version_number", "") or 0
    if int(version) >= 0x00030600:
      if (not weechat.config_is_set_plugin('lag_warning') or weechat.config_get_plugin('lag_warning') == 'on'):
        weechat.prnt('', '%s%sWARNING! This %s script is currently rather inefficient, and requires a rewrite.' % (weechat.prefix('error'), weechat.color('red'), SCRIPT_NAME))
        weechat.prnt('', '%s  It may cause nasty lag when there are a lot of concurrent joins on the channel, as it evaluates the nicklist on every join.' % weechat.prefix('notice'))
        weechat.prnt('', '%s  For servers like twitch.tv and bitlbee, you might want to exclude the server with the new setting:' % weechat.prefix('notice'))
        weechat.prnt('', '%s  %s  /set plugins.var.python.clone_scanner.hooks.excluded_servers twitchtv,bitlbee' % (weechat.prefix("notice"), weechat.color('*white')))
        weechat.prnt('', '%s  This script update is an emergency update to add the above option, and warn the users of this script.' % weechat.prefix("notice"))
        weechat.prnt('', '%s  You can disable this warning with:%s /set plugins.var.python.clone_scanner.lag_warning off' % (weechat.prefix("notice"), weechat.color('*white')))

      cs_set_default_settings()
      cs_buffer = weechat.buffer_search("python", "clone_scanner")

      weechat.hook_config( 'plugins.var.python.%s.*' % SCRIPT_NAME, 'toggle_refresh', '' )
      add_hooks()

      weechat.hook_command(SCRIPT_COMMAND,
                            SCRIPT_DESC,
                            "[scan] [[plugin.][network.]channel] | [advertise] | [help]",
                            "the target_buffer can be: \n"
                            "- left out, so the current channel buffer will be scanned.\n"
                            "- a plain channel name, such as #weechat, in which case it will prefixed with the current network name\n"
                            "- a channel name prefixed with network name, such as Freenode.#weechat\n"
                            "- a channel name prefixed with plugin and network name, such as irc.freenode.#weechat\n"
                            "See /set plugins.var.python.clone_scanner.* for all possible configuration options",

                            " || scan %(buffers_names)"
                            " || advertise"
                            " || help",

                            "cs_command_main", "")
  else:
    weechat.prnt("","%s%s %s" % (weechat.prefix("error"),SCRIPT_NAME,": needs version 0.3.6 or higher"))
    weechat.command("","/wait 1ms /python unload %s" % SCRIPT_NAME)



"""
Copyright (c) 2014-2018 by Vlad Stoica <stoica.vl@gmail.com>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

History
-------
01-05-2014 - Vlad Stoica
Uses a sqlite3 database to store triggers with the replies, has the
ability to ignore channels.
16-08-2015 - Vlad Stoica
Fixed a bug where replies couldn't have `:' in them.
15-02-2018 - Vlad Stoica
Added regex support in triggers, and edited syntax of adding triggers.
The command is now 'add "trigger" "reply"'. Quote marks can be escaped
in triggers or replies by prefixing them with a backslash ('\'). For
example, 'add "\"picture\"" "say \"cheese\"!"' is a valid command and
will reply with 'say "cheese"!' whenever it finds '"picture"' sent.
29-04-2020 - Fisher
Some new functions:
      - multiple matches ("match1|match2|match3" etc)
      - random selected multiple replites ("reply1|reply2|reply3" etc)
      - can ignore nicks (and ignore itself) for example bots
      - matches now case insensitive
      - utf-8 added
      - cooldown (max. n replies in t time)
      - random delay, so more human-like
      - even more randomness: can specify randomness of the reply/replies group.
2020-09-05 - new function - actions
2021-05-06 - Sébastien Helleu <flashcode@flashtux.org>
Add compatibility with WeeChat >= 3.2 (XDG directories).

2022-07-22 - Nils Görs (libera.#weechat)
      - fix bug: https://github.com/weechat/scripts/issues/459
      - add autocompletion
      - add option for sqlite3 filename
      - help text will be displayed using /help command

Bugs: not that i'm aware of.
"""

try:
    import weechat
    import sqlite3
    import re
    import random
    import sys
except ImportError:
    raise ImportError("Failed importing weechat, sqlite3, re or random")
import os

SCRIPT_NAME = "triggerreply"
SCRIPT_AUTHOR = "Vlad Stoica <stoica.vl@gmail.com>"
SCRIPT_VERSION = "0.4.4"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC = "Auto replies when someone sends a specified trigger. Now with 100% more regex!"
pcooldown  = 1
""" This is all I need so far :) """
colorcodes = { "^Cb":"\x02","^CR":"\x0F","^Ci":"\x1D" }

def cooldown_timer_cb(data, remaining_calls):
    global pcooldown
    if ( pcooldown > 0 ):
         pcooldown -= 1
    return weechat.WEECHAT_RC_OK

def print_help():
    weechat.prnt('','%s%s %s %s' % (weechat.prefix('error'),SCRIPT_NAME,': see /help',SCRIPT_NAME))
    return weechat.WEECHAT_RC_OK

def debug(mlevel, message):
    if int(weechat.config_get_plugin('debug')) >= int(mlevel):
       weechat.prnt("", "DEBUG: %s" % message)

def create_db(delete=False):
    debug(3, "Creating basic database.")
    """ create the sqlite database """
    if delete:
        os.remove(db_file)
    temp_con = sqlite3.connect(db_file)
    cur = temp_con.cursor()
    cur.execute("CREATE TABLE triggers(id INTEGER PRIMARY KEY, trig VARCHAR, reply VARCHAR, prob INTEGER);")
    cur.execute("INSERT INTO triggers(trig, reply, prob) VALUES ('trigge.rs', 'Automatic reply', '1');")
    cur.execute("CREATE TABLE banchans(id INTEGER PRIMARY KEY, ignored VARCHAR);")
    cur.execute("INSERT INTO banchans(ignored) VALUES ('rizon.#help');")
    cur.execute("CREATE TABLE ignorenicks(id INTEGER PRIMARY KEY, ignored VARCHAR);")
    cur.execute("INSERT INTO ignorenicks(ignored) VALUES ('dumanet.#DumaNet.Neo');")
    temp_con.commit()
    cur.close()



def check_db():
    temp_con = sqlite3.connect(db_file)
    cur = temp_con.cursor()

    try:
        """ Try to add record enchated with probability """
        cur.execute("INSERT INTO triggers(trig, reply, prob) VALUES (?,?,?)", ('JJORAIGPADMLOLYUGSBZ',"",1))
    except:
        """ If it fails, hope the best and assume it is just an older schema """
        cur.execute("ALTER TABLE triggers ADD COLUMN prob INTEGER")

    """ Clean up the mess """
    cur.execute("DELETE FROM triggers WHERE trig='JJORAIGPADMLOLYUGSBZ'")
    temp_con.commit()
    cur.close()



def search_trig_cb(data, buf, date, tags, displayed, highlight, prefix, message):
    """ function for parsing sent messages """
    global pcooldown

    """ Prevent infinite loop/flood, no more messages than n (approx 3) in 300 secs """
    if ( pcooldown > 300 ): return weechat.WEECHAT_RC_OK

    """ Save some CPU cycles """
    if (prefix == '-->' or prefix == '<--' or prefix == '--' or prefix == ' *' or prefix == ""): return weechat.WEECHAT_RC_OK

    bufname = weechat.buffer_get_string(buf, "name")

    if bufname == 'weechat': return weechat.WEECHAT_RC_OK

    """ Ignore myself """
    mynick =  weechat.buffer_get_string(buf, "localvar_nick")
    if re.search('[@+~]?' + mynick, prefix):
        """ weechat.prnt("", "Ignored myself.") """
        return weechat.WEECHAT_RC_OK


    database = sqlite3.connect(db_file)
    cursor = database.cursor()
    pure = weechat.string_remove_color(message,"")

    debug(1, "Nick in question:'%s" % bufname + '.' + prefix.translate(None,'@+~') + "'")

    for row in cursor.execute("SELECT ignored from ignorenicks;"):
        if re.search(row[0], bufname + '.' + prefix.translate(None,'@+~')):
            """ weechat.prnt("", "Nick ignored: %s" % row[0]) """
            return weechat.WEECHAT_RC_OK

    for row in cursor.execute("SELECT ignored from banchans;"):
        if  bufname == row[0]:
            return weechat.WEECHAT_RC_OK

    for row in cursor.execute("SELECT * FROM triggers"):
        delay = random.randint(4,9)

        pattern = row[1].encode('utf8')
        pattern = pattern.replace("%N", mynick)
        replydata = row[2].encode('utf8')
        prob = int(row[3])

        for ccode, chex in list(colorcodes.items()):
            replydata = replydata.replace(ccode,chex)

        try:
            nick = re.sub('^[+%@]','', prefix)
            debug(2, "prefix: %s, mynick: %s, nick: %s, pattern: %s, prob: %s, pure: %s" % (prefix, mynick, nick, pattern, str(prob), pure))

            r = re.compile(pattern,re.I | re.U)

            if r.search(pure) is not None:
                weechat.prnt("", "Matched")

                """ Meh, not really sure how random it is, but probably good enough """
                if ( prob > 1 and random.randint(1,prob) == 1):
                    debug(1, "Randomly ignored.")
                    return weechat.WEECHAT_RC_OK

                weechat.prnt("", "Match: %s" % r.search(pure).group(0))
                myreply = "n/a"
                if prob < 0:
                    """ -1 means this is action, not saying """
                    delay = 0
                    debug(1,"Command mode triggered.")
                    infolist = weechat.infolist_get("irc_nick", "", bufname.replace(".",","))
                    while weechat.infolist_next(infolist):
                       _nick = weechat.infolist_string(infolist, 'name')
                       if _nick == nick:
                          hostinfo = weechat.infolist_string(infolist,'host')
                          break
                    mask = hostinfo.split('@')[1]
                    weechat.prnt("", "mask: %s" % mask)
                    weechat.infolist_free(infolist)

                    for myreply in replydata.split('|'):
                        myreply = myreply.replace("%n", nick)
                        myreply = myreply.replace("%N", mynick)
                        myreply = myreply.replace("%m", mask)
                        myreply = myreply.replace("%c", bufname.split(".")[1])
                        weechat.prnt("", "Command: %s" % myreply)
                        if delay > 0:
                            weechat.command(buf, "/wait %s %s" % (delay, myreply))
                        else:
                            weechat.command(buf, "%s" % myreply)
                        delay++2

                    return weechat.WEECHAT_RC_OK

                myreply = random.choice(replydata.split('|'))
                myreply = myreply.replace("%n", nick)
                weechat.prnt("", "reply: %s" % myreply)
                weechat.prnt("", "%s triggered." % pattern)
                weechat.command(buf, "/wait %s /say %s" % (delay, myreply))
                pcooldown += 100
        except:
            weechat.prnt("", "NOMatch")
            if pattern == pure:
                weechat.command(buf, "/wait %s /say %s" % (delay, myreply))
                pcooldown += 120

    return weechat.WEECHAT_RC_OK


def command_input_callback(data, buffer, argv):
    """ function called when `/triggerreply args' is run """
    database = sqlite3.connect(db_file)
    cursor = database.cursor()
    command = argv.split()

    if len(command) == 0:
        return weechat.WEECHAT_RC_ERROR

    if command[0] == "list":
        weechat.prnt("", "List of triggers with replies:")
        for row in cursor.execute("SELECT * FROM triggers;"):
            weechat.prnt("", (str(row[0]) + ". " + str(row[1]) + " -> " + str(row[2]) + "  [Prob: " + str(row[3]) + "]"))
#            weechat.prnt("", str(row[0]) + ". " + row[1].encode('utf8') + " -> " + row[2].encode('utf8') + "  [Prob: " + str(row[3]) + "]")

        weechat.prnt("", "\nList of ignored channels:")
        for row in cursor.execute("SELECT ignored FROM banchans;"):
            weechat.prnt("", row[0])

        weechat.prnt("", "\nList of ignored nicks:")
        for row in cursor.execute("SELECT ignored FROM ignorenicks;"):
            weechat.prnt("", str(row[0]))

    elif command[0] == "add":
        if len(argv) == len(command[0]):
            print_help()
            return weechat.WEECHAT_RC_ERROR

        if argv.count('"') < 4:
            print_help()
            return weechat.WEECHAT_RC_ERROR

        pos = []
        for k, v in enumerate(argv):
            if v == '"' and argv[k - 1] != '\\':
                pos.append(k)

        if (len(pos) != 6 and len(pos) != 4):
            print_help()
            return weechat.WEECHAT_RC_ERROR

        trigger = argv[pos[0] + 1:pos[1]].replace('\\"', '"')
        reply = argv[pos[2] + 1:pos[3]].replace('\\"', '"')

        prob = 1
        if (len(pos) == 6):
            prob = int(argv[pos[4] + 1:pos[5]])

        try:
            cursor.execute("INSERT INTO triggers(trig, reply, prob) VALUES (?,?,?)", (trigger, reply, prob))
#            cursor.execute("INSERT INTO triggers(trig, reply, prob) VALUES (?,?,?)", (trigger.encode('utf8'), reply.encode('utf8'), prob))
#            cursor.execute("INSERT INTO triggers(trig, reply, prob) VALUES (?,?,?)", (trigger.decode('utf8'), reply.decode('utf8'), prob))
        except:
            print_help()
            weechat.prnt("", "DB Insert error.")
            return weechat.WEECHAT_RC_ERROR

        database.commit()
        weechat.prnt("", "Trigger added successfully!")
    elif command[0] == "remove":
        if len(argv) == len(command[0]):
            print_help()
            return weechat.WEECHAT_RC_ERROR

        try:
            cursor.execute("DELETE FROM triggers WHERE id = ?", (argv[7:],))
        except:
            print_help()
            return weechat.WEECHAT_RC_ERROR

        database.commit()
        weechat.prnt("", "Trigger successfully removed.")
    elif command[0] == "ignore":
        if len(argv) == len(command[0]):
            print_help()
            return weechat.WEECHAT_RC_ERROR

        try:
            cursor.execute("INSERT INTO banchans(ignored) VALUES (?)", (command[1],))
        except:
            print_help()
            return weechat.WEECHAT_RC_ERROR

        database.commit()
        weechat.prnt("", "Channel successfully added to ignore list!")
    elif command[0] == "parse":
        if len(argv) == len(command[0]):
            print_help()
            return weechat.WEECHAT_RC_ERROR

        try:
            cursor.execute("DELETE FROM banchans WHERE ignored = ?", (command[1],))
        except:
            print_help()
            return weechat.WEECHAT_RC_ERROR

        database.commit()
        weechat.prnt("", "Channnel being watched again.")

    elif command[0] == "ignorenick":
        if len(argv) == len(command[0]):
            print_help()
            return weechat.WEECHAT_RC_ERROR

        try:
            cursor.execute("INSERT INTO ignorenicks(ignored) VALUES (?)", (command[1],))
        except:
            print_help()
            return weechat.WEECHAT_RC_ERROR

        database.commit()
        weechat.prnt("", "Nick successfully added to ignore list!")
    elif command[0] == "watchnick":
        if len(argv) == len(command[0]):
            print_help()
            return weechat.WEECHAT_RC_ERROR

        try:
            cursor.execute("DELETE FROM ignorenicks WHERE ignored = ?", (command[1],))
        except:
            print_help()
            return weechat.WEECHAT_RC_ERROR

        database.commit()
        weechat.prnt("", "Nick successfully removed from ignored.")

    return weechat.WEECHAT_RC_OK


if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""):
    options = {
        'directory': 'data',
    }

    if weechat.config_get_plugin('database') == "":
        weechat.config_set_plugin('database', "%h/trigge.rs")
        db_file = weechat.string_eval_path_home("%h/trigge.rs", {}, {}, options)
    else:
       db_file = weechat.string_eval_path_home(weechat.config_get_plugin('database'), {}, {}, options)

    if weechat.config_get_plugin('debug') == "":
        weechat.config_set_plugin('debug', "0")

    random.seed()

    if not os.path.isfile(db_file):
        create_db()

    check_db()

    weechat.hook_print("", "", "", 1, "search_trig_cb", "")
    weechat.hook_command(SCRIPT_NAME, SCRIPT_DESC, 'Triggerreply (trigge.rs) script. Automatically replies over specified triggers.\n'
'Usage: /triggerreply [list | add | remove | ignore | parse] ARGUMENTS\n\n'
'Commands:\n'
'    list   - lists the triggers with replies, and ignored channels\n'
'    add    - three arguments: "trigger", "reply" and probability\n'
'           - adds a trigger with the specified reply and probability\n'
'           - probability 1 = 1/1 (100%), 5 = 1/5 (20 %) - optional, default is 1 (100%)\n'
'           - negative probability means action, see examples\n'
'           - %n in the reply will be replaced by the nick of the matching line\n'
'           - %N replaced by "my" nick\n'
'           - %m replaced by host and mask *!*@\n'
'           - %c replaced by channel name\n\n'
'    remove - one argument: "trigger"\n'
'           - remove a trigger\n'
'    ignore - one argument: "server.#channel"\n'
'           - ignores a particular channel from a server\n'
'    parse  - one argument: "server.#channel"\n'
'           - removes a channel from ignored list\n'
'ignorenick - one argument: "server.#channel.Nick"\n'
'           - ignores a particular nick from a server.#channel\n'
'watchnick  - one argument: "server.#channel.Nick"\n'
'           - removes a nick from ignored list\n\n'
'Examples:\n'
'    /triggerreply add "^H(i|ello|ey)[ .!]*" "Hey there!|Hi matey|Aloha!" "1"\n'
'    /triggerreply add "lol" "not funny tho" "5"\n'
'    /triggerreply remove 2\n'
'    /triggerreply ignore rizon.#help\n'
'    /triggerreply parse rizon.#help\n'
'    /triggerreply ignore rizon.#help.\n'
'    /triggerreply ignorenick rizon.#help.Bot\n'
'    /triggerreply watchnick rizon.#help.Bot\n\n'
'Auto greetings:\n'
'/triggerreply add "(hi|hello|hey|howdy)[,: ]+%N" "Hi, %n.|Hello, %n." "1"\n'
'/triggerreply add "%N[,: ]+(hi|hello|hey|howdy)" "Hi, %n.|Helllo, %n." "1"\n\n\n'
'Kick on adult content. Probability -1 means the strings between | are command executed in order:\n'
'/triggerreply add "https?://(www\.)?pornhub\.com|https?://(www\.)?xhamster\.com" "/msg chanserv op %c %N|/kick %n No adult content here, bye|/ban *!*@%m|/msg chanserv deop %c %N" "-1"', "", "list||add||remove||ignore||parse||ignorenick||watchnick",
                         "command_input_callback", "")

    """ fire every sec """
    hook = weechat.hook_timer(1000, 0, 0, "cooldown_timer_cb", "")



"""
Copyright (c) 2021-present Mohan Raman <mohan43u@gmail.com>.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

   * Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above
copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the
distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

2021-07-01: Mohan R
     0.0.1: Initial release with IrssiNotifier
            (https://irssinotifier.appspot.com/)

2024-10-06: Mohan R
     0.1.0: switched to gotify (https://gotify.net/)
            it should also support ntfy.sh (https://ntfy.sh)

Description: notifier for push notification
Project URL: https://gitlab.com/mohan43u/weenotifier
"""

from urllib.parse import urlencode

try:
    import weechat
    result = weechat.register("weenotifier",
                              "Mohan R",
                              "0.1.0",
                              "BSD 2-Clause Simplified",
                              "notifier for push notification",
                              "shutdown_cb",
                              "")
except ImportError as exception:
    raise ImportError("This script has to run under" +
                      "WeeChat (https://weechat.org/)") from exception


class WeeNotifier:
    """
    Weenotifier - primary class.
    """

    def __init__(self):
        """
        Weenotifier - initialization.
        """

        self.options = {
            "url": "",  # gotify: https://selfhostedgotify.yrdomain.tld/message
                        # ntfy.sh: https://ntfy.sh/youruniquetopic
            "token": ""
        }

        for option, value in self.options.items():
            if not weechat.config_is_set_plugin(option):
                _ = weechat.config_set_plugin(option,
                                              value)

        self.url = weechat.config_get_plugin("url")
        if len(self.url) <= 0:
            raise NameError("weenotifier: 'url' not configured")

        # for gotify
        token = weechat.config_get_plugin("token")
        if len(token) > 0:
            self.url += "?token=" + token

        self.version = weechat.info_get("version_number",
                                        "") or "0"
        _ = weechat.hook_print("",
                               "",
                               "",
                               1,
                               "message_cb",
                               "")

    def message(self,
                _data: str,
                buffer: str,
                _date: int,
                tags: list[str],
                _displayed: int,
                highlight: int,
                prefix: str,
                message: str):
        """
        Send message to push notification server.
        """

        if highlight or ("notify_private" in tags):
            channel = weechat.buffer_get_string(buffer, "short_name") or \
                weechat.buffer_get_string(buffer, "name")
            message = prefix + ": " + message
            post = channel + ": " + message

            # format for gotify
            if "?token=" in self.url:
                post = urlencode({"title": channel,
                                  "message": message,
                                  "priority": "4"})

            _ = weechat.hook_process_hashtable("url:" + self.url,
                                               {"post": "1",
                                                "postfields": post},
                                               10000,
                                               "result_cb",
                                               "")
        return weechat.WEECHAT_RC_OK

    def result(self,
               data: str,
               url: str,
               returncode: int,
               output: str,
               err: str):
        """
        Print result of url request
        """

        if returncode != 0 or ("error" in output) or len(err) > 0:
            print(url,
                  data,
                  returncode,
                  output,
                  err)
            return weechat.WEECHAT_RC_ERROR
        return weechat.WEECHAT_RC_OK

    def shutdown(self):
        """
        Shutdown
        """

        print("shutdown invoked")
        return weechat.WEECHAT_RC_OK


weenotifier: WeeNotifier | None = None


def message_cb(data: str,
               buffer: str,
               date: int,
               tags: list[str],
               displayed: int,
               highlight: int,
               prefix: str,
               message: str):
    """
    Message callback by weechat-C-api.
    """

    if weenotifier is not None:
        return weenotifier.message(data,
                                   buffer,
                                   date,
                                   tags,
                                   displayed,
                                   highlight,
                                   prefix,
                                   message)
    return weechat.WEECHAT_RC_ERROR


def result_cb(data: str,
              url: str,
              returncode: int,
              output: str,
              err: str):
    """
    Result callback by weechat-C-api
    """

    if weenotifier is not None:
        return weenotifier.result(data,
                                  url,
                                  returncode,
                                  output,
                                  err)
    return weechat.WEECHAT_RC_ERROR


def shutdown_cb():
    """
    Shutdown callback by weechat-C-api
    """

    if weenotifier is not None:
        return weenotifier.shutdown()
    return weechat.WEECHAT_RC_ERROR


def main():
    global weenotifier
    weenotifier = WeeNotifier()


if __name__ == "__main__":
    main()



# coding=utf-8
# WeeChat script for nvim style digraphs. Enables entering math symbols easily.

# Copyright (C) 2023 narodnik <policeterror@dyne.org>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Main repository, version history:  https://github.com/narodnik/weechat-digraph-latex

import weechat

# Substrings surrounded by this will have replacement active
MODIFIER = "$"

weechat.register("digraph", "narodnik", "1.0", "GPL3",
                 "Digraphs like nvim for inputting math symbols",
                 "", "")

weechat.hook_modifier("input_text_display",    "modifier_cb", "")
weechat.hook_modifier("input_text_for_buffer", "modifier_cb", "")

sup_vals = "⁰¹²³⁴⁵⁶⁷⁸⁹ᵃᵇᶜᵈᵉᶠᵍʰⁱʲᵏˡᵐⁿᵒᵖʳˢᵗᵘᵛʷˣʸᶻ⁺⁻⁼⁽⁾"
sup_keys = "0123456789abcdefghijklmnoprstuvwxyz+-=()"

sub_vals = "₀₁₂₃₄₅₆₇₈₉₊₋₌₍₎"
sub_keys = "0123456789+-=()"

symbols = [
    ("ZZ", "ℤ"),
    ("QQ", "ℚ"),
    ("FF", "𝔽"),
    ("a:", "𝔞"),
    ("b:", "𝔟"),
    ("c:", "𝔠"),
    ("p:", "𝔭"),
    ("in", "∈"),
    ("ni", "∉"),
    ("(_", "⊆"),
    ("(<", "⊊"),
    ("(!", "⊈"),
    (":.", "·"),
    (".,", "…"),
    (".3", "⋯"),
    ("**", "×"),
    ("i8", "∞"),
    ("</", "⟨"),
    ("/>", "⟩"),
    ("ff", "ϕ"),
    ("=>", "⇒"),
    ("==", "⇔"),
    ("->", "→"),
    ("TE", "∃"),
    ("!=", "≠"),
    ("=3", "≡"),
    ("=<", "≤"),
    ("<=", "≤"),
    (">=", "≥"),
    ("=?", "≌"),
    ("RT", "√"),
    ("(U", "∩"),
]

greek_key = "abcdefghiklmnopqrstuwxyz"
greek_cap = "ΑΒΞΔΕΦΓΘΙΚΛΜΝΟΠΨΡΣΤΥΩΧΗΖ"
greek_min = "αβξδεφγθικλμνοπψρστυωχηζ"

START_EXPR = 1
END_EXPR = 2

def build_replacement_table():
    table = []
    table.extend(symbols)
    # Superscript
    for key, val in zip(sup_keys, sup_vals):
        table.append((f"^{key}", val))
    # Subscript
    for key, val in zip(sub_keys, sub_vals):
        table.append((f"_{key}", val))
    # Greek letters
    assert greek_key == greek_key.lower()
    for key, cap, min in zip(greek_key, greek_cap, greek_min):
        table.append((f"{key.upper()}*", cap))
        table.append((f"{key}*", min))
    return table

replacement_table = build_replacement_table()

def find_all(a_str, sub):
    start = 0
    while True:
        start = a_str.find(sub, start)
        if start == -1: return
        yield start
        start += len(sub) # use start += 1 to find overlapping matches

def replace_symbols(a_str):
    for key, value in replacement_table:
        a_str = a_str.replace(key, value)
    return a_str

def lexer(string):
    tokens = []
    last_pos = 0
    for pos in find_all(string, MODIFIER):
        tokens.append(string[last_pos:pos - len(MODIFIER) + 1])
        tokens.append(MODIFIER)
        last_pos = pos + len(MODIFIER)
    if last_pos < len(string):
        tokens.append(string[last_pos:])

    active = False
    for i, token in enumerate(tokens):
        if token == MODIFIER:
            if not active:
                tokens[i] = START_EXPR
            else:
                tokens[i] = END_EXPR
            active = not active

    return tokens

def compile(tokens):
    state = END_EXPR
    result = ""
    current_word = None
    is_last = lambda i: i == len(tokens) - 1
    for i, token in enumerate(tokens):
        if token == START_EXPR:
            assert state == END_EXPR
            result += current_word
            current_word = None
            # When at the very last token, keep the unclosed MODIFIER
            if is_last(i):
                result += MODIFIER
            # Now change state to open expr
            state = START_EXPR
        elif token == END_EXPR:
            assert state == START_EXPR
            # Close the prev expr
            result += replace_symbols(current_word)
            current_word = None
            # Re-open normal mode
            state = END_EXPR
        else:
            assert state in (START_EXPR, END_EXPR)
            assert current_word is None
            current_word = token
    if current_word is not None:
        if state == START_EXPR:
            result += MODIFIER
            result += replace_symbols(current_word)
        else:
            assert state == END_EXPR
            result += current_word
    return result

def modifier_cb(data, modifier, modifier_data, string):
    tokens = lexer(string)
    result = compile(tokens)
    return result




# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2018 by Simmo Saan <simmo.saan@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# History:
#
# 2018-06-19, Simmo Saan <simmo.saan@gmail.com>
#   version 0.5: ignore channel modes arguments
# 2015-08-07, Simmo Saan <simmo.saan@gmail.com>
#   version 0.4: option for invite-only channels
# 2015-08-07, Simmo Saan <simmo.saan@gmail.com>
#   version 0.3: options to control risky channel cycling
# 2015-07-03, Simmo Saan <simmo.saan@gmail.com>
#   version 0.2: ability to rejoin passworded channels
# 2015-07-01, Simmo Saan <simmo.saan@gmail.com>
#   version 0.1: initial script
#

"""
Force nick change on channels which disallow it
"""

from __future__ import print_function

SCRIPT_NAME = "force_nick"
SCRIPT_AUTHOR = "Simmo Saan <simmo.saan@gmail.com>"
SCRIPT_VERSION = "0.5"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC = "Force nick change on channels which disallow it"

IMPORT_OK = True

try:
    import weechat
except ImportError:
    print("This script must be run under WeeChat.")
    print("Get WeeChat now at: http://www.weechat.org/")
    IMPORT_OK = False

SETTINGS = {
    "cycle_detach": (
        "off",
        "automatically cycle channels which are not open in WeeChat"),
    "cycle_key": (
        "on",
        "automatically cycle passworded channels (+k)"),
    "cycle_invite": (
        "off",
        "automatically cycle invite-only channels (+i)")
}

import re

servers = {}

def parse_message(signal_data):
    hashtable = weechat.info_get_hashtable("irc_message_parse", {"message": signal_data})

    # parse arguments string into usable pieces
    args = hashtable["arguments"].split(":", 1)
    hashtable["args"] = args[0].split()
    if len(args) > 1:
        hashtable["text"] = args[1]

    return hashtable

def channel_block(server, channel):
    fail = None
    config_cycle = lambda opt: weechat.config_string_to_boolean(weechat.config_get_plugin("cycle_%s" % opt))

    channels = weechat.infolist_get("irc_channel", "", "%s,%s" % (server, channel))
    if weechat.infolist_next(channels):
        modes = weechat.infolist_string(channels, "modes")
        if " " in modes:
            modes, modes_args = modes.split(" ", 1)

        if not config_cycle("key") and weechat.infolist_string(channels, "key") != "":
            fail = "cycle_key"
        elif not config_cycle("invite") and "i" in modes:
            fail = "cycle_invite"
    elif not config_cycle("detach"):
        fail = "cycle_detach"

    weechat.infolist_free(channels)

    if fail:
        weechat.prnt("", "%s: won't automatically cycle %s.%s: %s" % (SCRIPT_NAME, server, channel, fail))
    else:
        servers[server]["channels"].append(channel)
        buffer = weechat.buffer_search("irc", server)
        weechat.command(buffer, "/part %s" % channel)
        weechat.command(buffer, "/nick %s" % servers[server]["nick"])

def nick_out_cb(data, signal, signal_data):
    server = signal.split(",")[0]
    parsed = parse_message(signal_data)
    nick = parsed["args"][0]

    if server not in servers: # initialize new nickchange
        servers[server] = {}
        servers[server]["channels"] = []

    servers[server]["nick"] = nick

    return weechat.WEECHAT_RC_OK

def nick_in_cb(data, signal, signal_data):
    server = signal.split(",")[0]
    parsed = parse_message(signal_data)
    mynick = weechat.info_get("irc_nick", server)

    if parsed["nick"] == mynick: # nick change worked
        channels = weechat.infolist_get("irc_channel", "", server)
        keys = {}
        while weechat.infolist_next(channels):
            keys[weechat.infolist_string(channels, "name")] = weechat.infolist_string(channels, "key")

        buffer = weechat.buffer_search("irc", server)
        for channel in servers[server]["channels"]:
            weechat.command(buffer, "/join -noswitch %s %s" % (channel, keys.get(channel, "")))

        weechat.infolist_free(channels)

        servers.pop(server)

    return weechat.WEECHAT_RC_OK

def unreal_cb(data, signal, signal_data):
    server = signal.split(",")[0]
    parsed = parse_message(signal_data)

    match = re.match(r"Can not change nickname while on (#\w+) \(\+N\)", parsed["text"])
    if match:
        channel = match.group(1)
        channel_block(server, channel)

    return weechat.WEECHAT_RC_OK

def freenode_cb(data, signal, signal_data):
    server = signal.split(",")[0]
    parsed = parse_message(signal_data)

    channel_block(server, parsed["args"][2])

    return weechat.WEECHAT_RC_OK

if __name__ == "__main__" and IMPORT_OK:
    if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""):
        weechat.hook_signal("*,irc_out1_nick", "nick_out_cb", "")
        weechat.hook_signal("*,irc_in_nick", "nick_in_cb", "")
        weechat.hook_signal("*,irc_in_447", "unreal_cb", "")
        weechat.hook_signal("*,irc_in_435", "freenode_cb", "")

        for option, value in SETTINGS.items():
            if not weechat.config_is_set_plugin(option):
                weechat.config_set_plugin(option, value[0])

            weechat.config_set_desc_plugin(option, "%s (default: \"%s\")" % (value[1], value[0]))
